
# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.012.
use strict;
use warnings;

use Module::Build 0.28;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.28"
  },
  "configure_requires" => {
    "Module::Build" => "0.28"
  },
  "dist_abstract" => "Mojo::Base + lexical \"has\"",
  "dist_author" => [
    "Adriano Ferreira <ferreira\@cpan.org>"
  ],
  "dist_name" => "Jojo-Base",
  "dist_version" => "0.7.0",
  "license" => "artistic_2",
  "module_name" => "Jojo::Base",
  "recommends" => {
    "Jojo::Role" => "v0.5.0"
  },
  "recursive_test_files" => 1,
  "requires" => {
    "Mojo::Base" => 0,
    "Mojo::Util" => 0,
    "Mojolicious" => "7.43",
    "Sub::Inject" => "v0.2.0",
    "experimental" => 0,
    "perl" => "5.018"
  },
  "test_requires" => {
    "ExtUtils::MakeMaker" => 0,
    "File::Spec" => 0,
    "FindBin" => 0,
    "Mojo::ByteStream" => 0,
    "Mojo::Collection" => 0,
    "Mojo::DOM" => 0,
    "Mojo::File" => 0,
    "Test::More" => "0.96"
  }
);


my %fallback_build_requires = (
  "ExtUtils::MakeMaker" => 0,
  "File::Spec" => 0,
  "FindBin" => 0,
  "Module::Build" => "0.28",
  "Mojo::ByteStream" => 0,
  "Mojo::Collection" => 0,
  "Mojo::DOM" => 0,
  "Mojo::File" => 0,
  "Test::More" => "0.96"
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
